/* PROJECT....:  2FILE.C
** FILE.......:  2file.c
** VERSION....:  1.0 (Turbo C++ 1.0 & MSC 5.0)
** AUTHOR.....:  Stephen D. Cooper
** NOTICE.....:  Copyright 1990 ZIFF Communications Co.
**
** NOTES......:  This file contains those routines that are unique to
**		 2FILE.EXE.
**
** COMPILE....:  Turbo C ==> tcc -D_2FILE_C_ -O 2file.c fcommon.c
**		 MSC     ==> cl /D_2FILE_C_ 2file.c fcommon.c
*/

/********************************************************************
**	INCLUDE FILES						   **
********************************************************************/

#define MAIN_FILE

#include <conio.h>
#include <dos.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys\stat.h>
#include "compiler.h"
#include "2fil_err.h"

#include "fcommon.h"

/********************************************************************
**      GLOBAL VARIABLES     				   	   **
********************************************************************/

unsigned long filesize = 0;		/* Size of the destination
					   file (if it exists) */
unsigned long bytes_needed;		/* size of the source
					   disk + header */

DISKSTATS dstats;			/* Disk statistics */

/********************************************************************
**	FUNCTION DECLARATIONS				   	   **
********************************************************************/

int ask_permission(void);
int check_dest_disk_size(int drvno);
int get_source_size(void);
int process_argc(int argc, char *argv[]);
int read_disk_write_file(void);
int write_header(FILE *fileptr);

void fanfare(void);
void help(void);

/********************************************************************
*********************************************************************
**	CODE STARTS  						   **
*********************************************************************
********************************************************************/

/********************************************************************
** FUNCTION....:  main()					   **
********************************************************************/

void main(int argc, char *argv[])
{
	int retval;			/* Value returned by funcs */

	/*
	** Install our ctrl-break handler
	** and our hardware error handler
	*/
	signal(SIGINT, ctrlbrk_handler);
	HARDERR(harderr_handler);

	fanfare();			/* Strut our stuff */

	/*
	** Process and error-check command line
	*/
	if((retval = process_args(argc, argv)) != 0)
		err_exit(retval);

	/*
	** Get the size of the source diskette
	*/
	if((retval = get_source_size()) != 0)
		err_exit(retval);

	/*
	** Check to see if the destination file exists.  If it does, get
	** the file's size (just in case the destination disk only has
	** enough room to continue processing if the file is deleted).
	*/
	if((retval = get_filesize(filename)) != 0)
		err_exit(retval);

	/*
	** Check available space on destination drive
	*/
	if((retval = check_dest_disk_size(*filename - 'A')) != 0)
		err_exit(retval);

	/*
	** Ask permission to continue
	*/
	if((retval = ask_permission()) != 0)
		err_exit(retval);

	/*
	** Main processing
	*/
	if((retval = read_disk_write_file()) != 0)
		err_exit(retval);

	/*
	** Clean-up and go home
	*/
	cputs("Processing completed\r\n\n");
	exit(0);
}

/********************************************************************
** FUNCTION..: ask_permission()				           **
********************************************************************/

int ask_permission(void)
{
	int retval = ERR_NO_ERRORS;

	printf("Source diskette drive:  %c: - ",
			dstats.drivenum + 'A');
	printf("(%u sides, %u tracks, %u sectors per track)\n",
			dstats.sides, dstats.numtracks,
			dstats.numsectors);
	printf("Destination file.....:  %s - ", filename);
	printf("(%lu bytes)\n\n", bytes_needed);

	printf("About to start processing.  ");

	if(continue_yn() == 'N')
		retval = ERR_USER_ABORT;

	printf("\n\n");

	return(retval);
}

/********************************************************************
** FUNCTION..: check_dest_disk_size() 			           **
********************************************************************/

int check_dest_disk_size(int drvno)
{
	unsigned long bytes_avail;
	union REGS inregs, outregs;

	/*
	** Find out how many bytes are available on the destination disk
	*/
	inregs.h.ah = DOSFUNC_GET_ALLOC_INFO;
	inregs.h.dl = drvno + 1;	 	/* 0=default, 1=A, ect. */

	do
	{
		harderr_status = HARDERR_OK;
		intdos(&inregs, &outregs);
	} while(harderr_status == HARDERR_RETRY);

	if(harderr_status == HARDERR_ABORT)
		return(ERR_USER_ABORT);

	if(outregs.x.ax == 0xffff)
		return(ERR_DEST_DRIVE_BAD);

	/*
	** Compute available destination disk bytes
	*/
	bytes_avail = (unsigned long)outregs.x.cx;
	bytes_avail *= (unsigned long)outregs.x.ax;
	bytes_avail *= (unsigned long)outregs.x.bx;
	bytes_avail += filesize;

	/*
	** Compare and exit with an error message if needed
	*/
	if(bytes_avail < bytes_needed)
	{
		printf("Bytes needed on drive %c: = %7lu\n", drvno + 'A',
				bytes_needed);
		printf("         Bytes available = %7lu\n\n", bytes_avail);

		return(ERR_DEST_DRIVE_SPACE);
	}

	return(ERR_NO_ERRORS);
}

/********************************************************************
** FUNCTION....:  fanfare() 				           **
********************************************************************/

void fanfare(void)
{
	cputs("\r\n2FILE version 1.0 Copyright 1990 ZIFF "
		  "Communications Co.\r\n");
	cputs("PC Magazine  Stephen D. Cooper.\r\n\n");
}

/********************************************************************
** FUNCTION..: get_filesize()				           **
********************************************************************/

int get_filesize(char *filename)
{
	int retval = ERR_NO_ERRORS;
	int statval;
	struct stat filestats;

	/*
	** 1.  See if there is a disk in the destination drive
	** 2.  If there is, see if destination file exists
	** 3.  If it does, get it's size
	*/
	do
	{
		harderr_status = HARDERR_OK;
		statval = stat(filename, &filestats);
	} while(harderr_status == HARDERR_RETRY);

	if(harderr_status == HARDERR_ABORT)
		return(ERR_USER_ABORT);

	/*
	** If the file exists, give the user a chance to abort
	*/
	if(statval == 0)
	{
		printf("Destination file (%s) already exists.  ", filename);
		printf("\n\nOverwrite the file (Y/N)? ");

		if(getyn() == 'Y')
			filesize = filestats.st_size;
		else
			retval = ERR_USER_ABORT;

		printf("\n\n");
	}
	return(retval);
}

/********************************************************************
** FUNCTION....:  get_source_size()			           **
********************************************************************/

int get_source_size(void)
{
	unsigned long drivesize;
	unsigned char far *media_id;
	union REGS inregs, outregs;
	struct SREGS sregs;

	inregs.h.ah = DOSFUNC_GET_DRIVE_DATA;
	inregs.h.dl = dstats.drivenum + 1;	/* 0=default, 1=A, etc. */

	do
	{
		harderr_status = HARDERR_OK;
		intdosx(&inregs, &outregs, &sregs);
	} while(harderr_status == HARDERR_RETRY);

	if(harderr_status == HARDERR_ABORT)
		return(ERR_USER_ABORT);

	/*
	** Fill in our disk structure with defaults
	*/
	dstats.sides = 2;
	dstats.numtracks = 40;
	dstats.numsectors = 9;
	dstats.sectorsize = outregs.x.cx;

	media_id = MK_FP(sregs.ds, outregs.x.bx);

	switch(*media_id)
	{
		case 0xf0:			/* 1.44MB 3.5 */
			dstats.numsectors = 18;
			dstats.numtracks = 80;
			break;
		case 0xf8:			/* Fixed disk -- ERROR */
			help();
			return(ERR_FIXED_DISK);
		case 0xf9:			/* 1.2M 5.25 or 720K 3.5 */
			drivesize = (unsigned long)outregs.x.cx;
			drivesize *= (unsigned long)outregs.h.al;
			drivesize *= (unsigned long)outregs.x.dx;
			if(drivesize > 737280L)
				dstats.numsectors = 15;
			dstats.numtracks = 80;
			break;
		case 0xfc:			/* 180KB 5.25 */
			dstats.sides = 1;
			break;
		case 0xfd:			/* 360KB 5.25 */
			break;
		case 0xfe:			/* 160KB 5.25 */
			dstats.numsectors = 8;
			dstats.sides = 1;
			break;
		case 0xff:			/* 320KB 5.25 */
			dstats.numsectors = 8;
			break;
		default:			/* Unknown disk size */
			return(ERR_UNKNOWN_FLOPPY);
	}

	/*
	** Calculate the number of bytes our destination file will take
	** The last line "+=" is our header size:  The struct + 3 initials
	*/
	bytes_needed = (unsigned long)dstats.sectorsize;
	bytes_needed *= (unsigned long)dstats.numsectors;
	bytes_needed *= (unsigned long)dstats.numtracks;
	bytes_needed *= (unsigned long)dstats.sides;
	bytes_needed += (unsigned long)sizeof(dstats) + 3;

	return(ERR_NO_ERRORS);
}

/********************************************************************
** FUNCTION..: help()					           **
********************************************************************/

void help(void)
{
	char spc = ' ';

	printf("Syntax:  2FILE drive filename\n\n");

	printf("Where:   drive    = The name of the floppy drive you ");
	printf("wish to copy.  This\n");
	printf("%19c must be either A: or B: and can be in either ", spc);
	printf("upper case\n");
	printf("%19c or lower case.\n\n", spc);

	printf("%8c filename = The name of the destination file.  ", spc);
	printf("If no drive letter is\n");
	printf("%19c given, the current drive will be used as the ", spc);
	printf("default.\n");
	printf("%19c The drive where filename will be written ", spc);
	printf("must be\n");
	printf("%19c different than the source drive ", spc);
	printf("(drive parameter).  If no\n");
	printf("%19c extension is given, .FLP will be used as the ", spc);
	printf("default.\n");
	printf("%19c The filename can be in either upper case or ", spc);
	printf("lower case.\n\n");
}

/*********************************************************************
** FUNCTION process_args()					    **
*********************************************************************/

int process_args(int argc, char *argv[])
{
	int retval;			/* Value returned by funcs */

	/*
	** Check the number of DOS command line arguments --
	** Must be the program name plus two arguments
	*/
	if((retval = check_numargs(argc, 3, 3)) != 0)
	{
		help();
		return(retval);
	}

	/*
	** Check the first parameter --
	** Must be "A:" or "B:" (case insignificant)
	*/
	retval = floppyletter_2_number(&dstats.drivenum, argv[1]);
	if(retval != 0)
	{
		help();
		return(retval);
	}

	/*
	** Check the second parameter -- Must be a filename
	** (default drive is current, default extension is .FLP)
	*/
	if((retval = check_filename(filename, argv[2], &dstats)) != 0)
	{
		help();
		return(retval);
	}

	return(ERR_NO_ERRORS);
}

/********************************************************************
** FUNCTION read_disk_write_file()				   **
********************************************************************/

int read_disk_write_file(void)
{
	int buffsize, retval = ERR_NO_ERRORS;
	int head, track;
	int xhead, xtrack, yloc;

#if !defined(TURBO_C)
	struct rccoord cursorpos;
#endif

	/*
	** Get our read/write buffer
	*/
	buffsize = dstats.sectorsize * dstats.numsectors;
	buffsize *= sizeof(char);
	if((read_buff = (unsigned char *)malloc(buffsize)) == 0)
		return(ERR_MEMORY);

	/*
	** Open the file
	*/
	do
	{
		harderr_status = HARDERR_OK;
		fileptr = fopen(filename, "wb");
	} while(harderr_status == HARDERR_RETRY);

	if(harderr_status == HARDERR_ABORT)
		return(ERR_USER_ABORT);

	if(fileptr == NULL)
		return(ERR_CANT_OPEN_OUTPUT);

	/*
	** At this point, it's not a good idea to allow the user to
	** recover from critical hardware errors.  For example, if the
	** user were to take out the current disk and insert a new one
	** and we tried to write to a file that doesn't exist, where
	** would we be?
	*/
	prg_status = HARDERR_CRITICAL;

	/*
	** Write our header to the file
	*/
	if((retval = write_header(fileptr)) != 0)
		return(retval);

	/*
	** Read floppy and write file
	*/
#if defined(TURBO_C)
	xhead = wherex() + 27;
	yloc = wherey();
#else
	cursorpos = _gettextposition();
	xhead = cursorpos.row + 27;
	yloc = cursorpos.col;
#endif
	xtrack = xhead + 9;
	printf("Reading and writing:  head 0, track 00");
	for(track = 0; track < dstats.numtracks; track++)
	{
		if(retval != ERR_NO_ERRORS)	/* Check for errors in */
			break;			/* previous passes */

		GOTOXY(xtrack, yloc);
		printf("%2u", track);
		for(head = 0; head < dstats.sides; head++)
		{
			GOTOXY(xhead, yloc);
			printf("%1u", head);
			retval = read_sectors(track, head, read_buff, &dstats);
			if(retval != ERR_NO_ERRORS)
				break;
			retval = fwrite(read_buff, buffsize, 1, fileptr);
			if(retval != 1)
			{
				retval = ERR_WRITING_FILE;
				break;
			}
			else
				retval = ERR_NO_ERRORS;
		}
	}

	cputs("\r\n\n");
	fclose(fileptr);
	free(read_buff);
	return(retval);
}

/********************************************************************
** FUNCTION write_header()					   **
********************************************************************/

int write_header(FILE *fileptr)
{
	char hdrinits[3];

	hdrinits[0] = INI_1;
	hdrinits[1] = INI_2;
	hdrinits[2] = INI_3;

	if(fwrite(hdrinits, sizeof(char) * 3, 1, fileptr) != 1)
		return(ERR_WRITING_FILE);

	if(fwrite(&dstats, sizeof(dstats), 1, fileptr) != 1)
		return(ERR_WRITING_FILE);

	return(ERR_NO_ERRORS);

}

/****************************
*****************************
**** END OF FILE 2file.c ****
*****************************
****************************/
