/* PROJECT....:  2FILE.C and 2FLOPPY.C
** FILE.......:  fcommon.h
** VERSION....:  1.0 (Turbo C++ 1.0 & MSC 5.0)
** AUTHOR.....:  Stephen D. Cooper
** NOTICE.....:  Copyright 1990 ZIFF Communications Co.
**
** NOTES......:  This header file contains constants, function
**		 declarations, and global variables that are used by
**		 both 2FILE.C and 2FLOPPY.C
*/

#if !defined(_FCOMMON_H_)
#define _FCOMMON_H_

/********************************************************************
**        #define CONSTANTS					   **
********************************************************************/

/*
** DOS and BIOS constants
*/
#define BIOS_DISK		   0x13		/* BIOS disk interrupt */
#define BIOSFUNC_FORMAT		   5		/* BIOS format track func */
#define BIOSFUNC_READ_SECTOR       2		/* BIOS Read sector func */
#define BIOSFUNC_RESET		   0		/* BIOS reset drive func */
#define BIOSFUNC_WRITE_SECTOR      3		/* BIOS write sector func */
#define DOSFUNC_GET_ALLOC_INFO     0x36		/* DOS get alloc info func */
#define DOSFUNC_GET_DRIVE_DATA     0x1c		/* DOS get drive data func */


/*
** File header initials
*/
#define INI_1 'P'				/* PC Magazine initials */
#define INI_2 'C'				/* Used in the file header */
#define INI_3 'M'

/*
** Hardware error codes
** Special note for HARDERR_CRITICAL, and HARDERR_NON_CRITICAL:
**		If it is safe to recover from a critical error, then the
**		global variable prg_status will be set to
**		HARDERR_NON_CRITICAL.  Otherwise it is set to
**		HARDERR_CRITICAL in which case, it is not safe to recover
**		from a critical error.  If this is the case and one occurs,
**		then the program displays a message, cleans-up and exits to
**		DOS.
*/
#define HARDERR_CRITICAL      0			/* Critical code sections */
#define HARDERR_NON_CRITICAL  1			/* Non-critical code */

#define HARDERR_OK	      0			/* Hardware error ok */
#define HARDERR_IGNORE 	      0			/* Crit. error ignore code */
#define HARDERR_RETRY         1			/* Crit. error retry code */
#define HARDERR_ABORT         2			/* Crit. error abort code */
#define HARDERR_MSC_RETN     21			/* MS-DOS error return */

#if defined(TURBO_C)
#define HARDERR_RET_CODE HARDERR_IGNORE
#else
#define HARDERR_RET_CODE HARDERR_MSC_RETN
#endif

/*
** Miscellaneous
*/
#if !defined(TRUE)
#define TRUE 1
#endif
#if !defined(FALSE)
#define FALSE 0
#endif

/********************************************************************
**         STRUCTURES       					   **
********************************************************************/

typedef struct { unsigned drivenum;		/* Drive #. 0=A, 1=B, etc. */
				 unsigned sides;/* Number of sides (heads) */
				 unsigned numtracks;	/* Tracks / side */
				 unsigned numsectors; 	/* Sectors / track */
				 unsigned sectorsize; 	/* Bytes / sector */
			   } DISKSTATS;

/********************************************************************
**        GLOBAL VARIABLES 					   **
********************************************************************/

#if defined(MAIN_FILE)
char filename[MAXPATH];				/* Output file name */
int harderr_status = HARDERR_OK;		/* Critical error status */
int prg_status = HARDERR_NON_CRITICAL;		/* Program status */
unsigned char *read_buff = NULL;		/* Read buffer */
FILE *fileptr;					/* I/O File pointer */
#else
extern char filename[];
extern int harderr_status;
extern int prg_status;
extern unsigned char *read_buff;
extern FILE *fileptr;
#endif

/********************************************************************
**     	FUNCTION DECLARATIONS				           **
********************************************************************/

int check_filename(char *destname, char *given, DISKSTATS *dstats);
int check_numargs(int argc, int minargs, int maxargs);
int continue_yn(void);
int floppyletter_2_number(unsigned *drivenum, char *drive);
int getyn(void);
int read_sectors(int track, int head, char *buff, DISKSTATS *dstats);
int reset_drive(int drivenumber);
int sector_read(union REGS *inregs, union REGS *outregs, char *buff,
				DISKSTATS *dstats);

void ctrlbrk_handler(void);
void err_exit(int errcode);

HARDERR_HANDLER;

#endif						/* #if !defined(_FCOMMON_H_) */
/******************************
*******************************
**** END OF FILE fcommon.h ****
*******************************
******************************/
